%%%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5
%DIFERENTIAL GAME 2 PLAYERS  NASH EQ
%(c) Jorge Herrera, Benjamin Ivorra, ngel Ramos  2018
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Vv1 Vv2 V1new V2new its]=RaBVItG(N,dt, verbose, tol)

%%%%%%%%%INPUTS
%
% N number of points for the state variables (X_1, X_2)
% dt time step for the Euler discretization

%%%%%%%%% THIS PROGRAM SOLVES THE DIFFERENTIAL GAME "TEST 3" FROM CACACE,
%%%%%%%%% Equilibria for a Class of Non-Cooperative Differential Games
%%%%%%%%% It can be easily adapted to solve a N-players Feedback Nash
%%%%%%%%% Differential Game


%%%%%%%%OUTPUTS
%Vv1 Value for player 1 interpolated in a 2D mesh
%Vv2 Value for player 2 interpolated in a 2D mes
%v1  Value player 1 in the original array
%v2  Value player 2 in the original array
%its iterations needed to converge

global sp gamma u01 u02 X1 X2 x01 x02 h Xtr  pold1 pold2 V1old V2old   disc   

sp = 1; %shape parameter RBF

h = dt;

disc = 1; %discount parameter
gamma = 1-(h*disc);

%Create 2 column vectors using halton distributed points between 0,1

[P] = CreatePoints(N,2,'h');
 
%reescale the points to the original domain

X1 = -2+(4)*P(:,1); %state variable Player 1
X2 = -2+(4)*P(:,2); %state variable Player 2

N2 = length(X1);

%Initialize controls to zero
u1old = zeros(N2,1);
u2old = zeros(N2,1);


%Initialize Value at algorithm step =0

V1old=zeros(N2,1);
V2old=zeros(N2,1);

%Generate a Matrix with state values in columns

Xtr=[X1 X2];

[n, d] = size(Xtr);
[ny, dy] = size(V1old);

 
%Module for the RBF
dist = pdist2(Xtr,Xtr);    
rbf = @(e,r) sqrt(1+(e*r).^2);
     
P = [ones(N,1) X1 X2 X1.^2 X2.^2 X1.*X2];
[filp, colp] = size(P)  ;
[filpp, colpp] = size(P');
 M = rbf( sp,dist);
 A = [M P ;P' zeros(filpp, colp)];
 V10 = [V1old;zeros(filpp,1)];
 V20 = [V2old;zeros(filpp,1)];
     
 
% Initial Values for parameters associated to RBF

pold1 = zeros(length(V10),1); 
pold2 = zeros(length(V20),1);



% Main Loop

its = 1;


dif = tol+1;


while dif>tol
for i=1:N2

    
    %for the selected points, iterate
    u01 = u1old(i);
    u02 = u2old(i);
    x01 = X1(i);
    x02 = X2(i);

    
    %game iteration LOOP
    tolg=0.00001;
    errorg=tolg+1;
    itsg=1;
    
    while (errorg(itsg)>tolg) && (itsg<5)
     itsg=itsg+1;
       
     options = optimset('Display', 'off');   
     [k11 val1] = fmincon(@funp1,u01,[],[],[],[],-1,1,[],options);
     [k22 val2] = fmincon(@funp2,u02,[],[],[],[],-1,1,[],options);
     policystar = [k11 k22];
     policyold = [u01 u02 ];
     errorg(itsg)= norm(policystar-policyold, inf);
     %It has, as hyperparameter a learning rate of 0.9
     u01=0.90*k11+0.10*u01;
     u02=0.90*k22+0.10*u02;
   
    end

   %update controls and values
    
   u1(i,1)=u01;
   u2(i,1)=u02;
   V1new(i,1)=funp1(u01);
   V2new(i,1)=funp2(u02);  

    
end


Vold=[V1old V2old];
Vnew=[V1new V2new];
pold=[pold1 pold2];


V10=[V1new;zeros(filpp,1)];
V20=[V2new;zeros(filpp,1)]; 
     

%update parameters for RBF

     pnew1 = A \ (V10); 
     pnew2 = A \ (V20);
  
pnew = [pnew1 pnew2];
pold1 = pnew1;
pold2 = pnew2;

%compute convergence
dif(its)=norm(Vnew-Vold,inf) 

%update values and controls
V1old=V1new;
V2old=V2new;
u1old=u1;
u2old=u2;
   

its=its+1; 


if verbose ==1

figure(1)
plot(dif)
drawnow

end
   
%From a Mesh Free, it builds a 2D representation of the solution using RBF    
   
points=N;
%%%generate a mesh

xx1=linspace(-2,2,points);
xx2=linspace(-2,2,points);
[X01 X02]=meshgrid(xx1,xx2);


%parameterize values

pv1 = A \ ([V1new;zeros(filpp,1)]); %- model.meanY);
pv2 = A \ ([V2new;zeros(filpp,1)]);
     

for i=1:length(xx1)
    for j=1:length(xx2)
        
    
    x01=X01(i,j);
    x02=X02(i,j);

    Xnew=[x01 x02];
    nq = size(Xnew, 1);
    
dist2=pdist2(Xnew,Xtr);

Pnew=[ones(1,1) Xnew(1) Xnew(2) Xnew(1).^2 Xnew(2).^2 Xnew(1).*Xnew(2)];
     [Nfilp, Ncolp]=size(Pnew)  ;
     [Nfilpp, Ncolpp]=size(Pnew')   ;
     

     M2=rbf( sp,dist2);
     AA2=[M2 Pnew];
     
  
Vv1(i,j)=AA2*pv1;
Vv2(i,j)=AA2*pv2;

   
    
end
end




if verbose == 2


xx1=linspace(-2,2,N);
xx2=linspace(-2,2,N);
[XX1,XX2]=meshgrid(xx1,xx2);

figure(2)
subplot(2,2,1)
surf(X01,X02,Vv1)
title('Value Player 1')
xlabel('X_1')
ylabel('X_2')
subplot(2,2,2)
surf(X01,X02,Vv2)
title('Value Player 2')
xlabel('X_1')
ylabel('X_2')
subplot(2,2,3)
plot(dif)
drawnow
% 
% 
% 
end



end





